﻿/*	File			CaseStudy_ENS_Server_CSharp.cs
	Purpose			Notification Service Server
	Author			Richard Anthony	(R.J.Anthony@gre.ac.uk)
	Date			February 2015
*/

using System;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Collections.Generic;

namespace ENS_Server_CSharp
{
    public partial class CaseStudy_ENS_Server_CSharp : Form
    {   // Define constants 
        public const string EVENT_NOTIFICATION_SERVER_NAME = "Event_NS";	// Event Notification Server
        public const Int32 MAX_EVENT_TYPE_NAME_LENGTH = 50;
        public const Int32 MAX_EVENT_VALUE_LENGTH = 50;
        public const Int32 DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH = 50;

        public const Int16 From_DIRECTORY_SERVICE_PORT = 8002;	// Fixed, defined by Directory Service
        public const Int16 ENS_APPLICATION_PORT = 8003;          // TCP, ENS_Server Binds to this port
        public const Int16 To_DIRECTORY_SERVICE_PORT = 8008;	    // Fixed, defined by Directory Service

        // Define communication types
        [StructLayout(LayoutKind.Sequential)]
        public struct DirectoryServiceRequest_PDU
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH)]
            public byte[] bBuffer;
        };

        [StructLayout(LayoutKind.Sequential)]
        public struct DirectoryServiceReply_PDU                 // Fixed, defined by Directory Service
        {   // IP address followed by port
            public byte a1;
            public byte a2;
            public byte a3;
            public byte a4;
            public uint iPort;
        };

        public enum EventNotificationService_Message_Type { SubscribeToEvent, UnSubscribeToEvent, PublishEvent, EventUpdate };

        // SOCKADDR_IN not defined in c# - define it and its constituent sub-structures
        // Needed because the ENS passes the message sender's address in this format
        [StructLayout(LayoutKind.Sequential)]
        public unsafe struct SockAddr
        {
            public ushort sa_family;
            public fixed byte sa_data[14]; // Note: use byte instead of char because sizeof(char) == 2 in C#
        }

        [StructLayout(LayoutKind.Sequential)]
        public unsafe struct SockAddr_In
        {
            public Int16 sin_family;
            public Int16 sin_port;
            public In_Addr sin_addr;
            public fixed byte sin_zero[8];
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct In_Addr
        {
            public byte s_b1, s_b2, s_b3, s_b4;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct EventNotificationService_Message_PDU
        {
            public byte iMessageType;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_EVENT_TYPE_NAME_LENGTH)]
            public byte[] cEventType;       //  (c++)  char[] cEventType[MAX_EVENT_TYPE_NAME_LENGTH];
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_EVENT_VALUE_LENGTH)]
            public byte[] cEventValue;      //  (c++)  char[] cEventValue[MAX_EVENT_VALUE_LENGTH];
            // In actual transmission, a 3-byte pad is placed here so that the SockAddr_In begins on a 4-byte boundary
            public SockAddr_In SendersAddress;
        };

        public const Int32 EventNotificationService_Message_PDU_Size = 120;
        
        public const Int32 MAX_CONNECTIONS = 30;
        public const Int32 MAX_EVENT_TYPES = 30;

        public struct Connection {
	        public Socket m_ConnectSock;
            public IPEndPoint m_ConnectIPEndPoint;
        	public bool bInUse;
        };

        public unsafe struct SubscriberListItem {
	        public int iConnectionIndex; // Index into connection array to identify the subscriber
        };

        public unsafe struct Event_Type { 
                // An event type is dynamically created whenever a publisher publishes a new event type or a consumer subscribes to a new event type
				// This is totally run-time configured, the designer does not know anything about the event types that may exist
				// Components are decoupled at the application level as well as at the communication level:
				//		- a producer can publish a new event type regardless of whether there is any consumer for that event-type
				//		- a consumer can subscribe to a new event type regardless of whether there is any producer for that event-type
	        public string sEventType;
	        public string sEventValue;
            public LinkedList<SubscriberListItem> SubscriberList; // Linked list of subscribers
            public IPEndPoint PublisherAddress;	// IP address and Port number of most recent publisher of this event type
	        public bool bInUse_ValueValid;			// Has the event been published
	        public bool bInUse_SubscriptionExists; // Is there at least one consumer waiting for this event
        };
     
 	    public Connection[] m_ConnectionArray;
        public Event_Type[] m_EventTypeArray;
        
        // Define member variables
        Socket m_RegisterSock;		// Socket for sending 'Register' messages to directory service
        Socket m_ListenSock;		// Socket for communication with clients
        IPEndPoint m_DirectoryService_IPEndPoint;   // for Broadcasting 'register' messages to the Directory server
        IPEndPoint m_localIPEndPoint;               // Address for Receiving client requests
        SockAddr_In m_LocalAddress_FormattedAs_SockAddr_In;
        System.Windows.Forms.Timer m_Registration_Timer;
        System.Windows.Forms.Timer m_ApplicationMessage_Timer;  // Needed because receives are non-blocking
 
        bool bRegisterAlreadyDone;
               
        public CaseStudy_ENS_Server_CSharp()
        {
            InitializeComponent();
	        m_ConnectionArray = new Connection[MAX_CONNECTIONS];
	        m_EventTypeArray = new Event_Type[MAX_EVENT_TYPES];
 	        bRegisterAlreadyDone = false;
            GetLocalHostAddressDetails();
            m_LocalAddress_FormattedAs_SockAddr_In = GetLocalIPAddress();
            CreateRegisterSocket();
            CreateListenSocket();

            // Initialise the connections array
	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
		        m_ConnectionArray[i].bInUse = false;
	        }	
	        // Initialise the EventType_List array
	        for(int i = 0; i < MAX_EVENT_TYPES; i++)
	        {
                m_EventTypeArray[i].bInUse_SubscriptionExists = false;
                m_EventTypeArray[i].bInUse_ValueValid = false;
                m_EventTypeArray[i].SubscriberList = new LinkedList<SubscriberListItem>();  // Linked list of subscribers
	        }	
        
	        Activate_ENS();
	        DisplayEventTypeList();	// Refresh the list of known event types
        }

        void CreateRegisterSocket()
        {   // UDP, non-blocking, broadcast
            try
            {
                m_RegisterSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_RegisterSock.Blocking = false;
                m_RegisterSock.EnableBroadcast = true;
                System.Net.IPAddress BroadcastIPAddress = System.Net.IPAddress.Broadcast;
                m_DirectoryService_IPEndPoint = new System.Net.IPEndPoint(BroadcastIPAddress, System.Convert.ToInt16(To_DIRECTORY_SERVICE_PORT));
                WriteEventLogEntry("m_RegisterSock Socket created");
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Create Socket failed", "ENS Server");
                CloseAndExit();
            }
            catch   // Catch any other exceptions (e.g. arising from creating the Endpoint)
            {
                MessageBox.Show("Invalid IP address or Port number", "ENS Server");
                CloseAndExit();
            }
        }

        void CreateListenSocket()
        {   // TCP, non-blocking
            try
            {
                m_ListenSock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ListenSock.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Create Socket failed", "ENS Server");
                CloseAndExit();
            }
            // Bind to the Application port that has been advertised via the Directory Service
            try
            {
                m_ListenSock.Bind(m_localIPEndPoint);
            }
            catch
            {
                MessageBox.Show("Bind failed", "ENS Server");
                CloseAndExit();
            }
            try
            {
                m_ListenSock.Listen(5);
            }
            catch
            {
                MessageBox.Show("Listen failed", "ENS Server");
                CloseAndExit();
            }
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   // Create the local IPEndPoint using the first IPv4 address and the pre-defined port
                    m_localIPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], ENS_APPLICATION_PORT);
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    string sStr = string.Format("Local Address: {0}.{1}.{2}.{3}  Port:{4}",
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), ENS_APPLICATION_PORT);
                    WriteEventLogEntry(sStr);
                    return;
                }
            }
            MessageBox.Show("Failed to get local address!", "ENS Server");
            CloseAndExit();
        }
       
        public SockAddr_In GetLocalIPAddress()
        {   // Get local IP address in SockAddr_In format for use in ENS application messages
            SockAddr_In SockAddr_In_LocalAddress = new SockAddr_In();
            byte[] bAddress = m_localIPEndPoint.Address.GetAddressBytes();
            SockAddr_In_LocalAddress.sin_addr.s_b1 = bAddress[0];
            SockAddr_In_LocalAddress.sin_addr.s_b2 = bAddress[1];
            SockAddr_In_LocalAddress.sin_addr.s_b3 = bAddress[2];
            SockAddr_In_LocalAddress.sin_addr.s_b4 = bAddress[3];
            SockAddr_In_LocalAddress.sin_port = ENS_APPLICATION_PORT;
            SockAddr_In_LocalAddress.sin_family = 2; // Equivallent to AF_INET in C++
            return SockAddr_In_LocalAddress;
        }

        void Activate_ENS()
        {
            SendRegisterMessage();              // Auto-register with Directory service immediately
            Start_Registration_Timer();         // Auto-register with Directory service periodically
            Start_ApplicationMessage_Timer();   // Deal with connection requests and received messages on short-periodic basis
        }

        void DisplayEventTypeList()
        {
            EventType_listBox.Items.Clear(); // Clear the Listbox
            for (int i = 0; i < MAX_EVENT_TYPES; i++)
            {
                if (true == m_EventTypeArray[i].bInUse_SubscriptionExists || true == m_EventTypeArray[i].bInUse_ValueValid)
                {
                    EventType_listBox.Items.Add(m_EventTypeArray[i].sEventType);
                }
            }
        }

        void SendRegisterMessage() // Broadcast message to Directory Server
        {
            String sRequest = "REGISTER:" + EVENT_NOTIFICATION_SERVER_NAME + ":" + ENS_APPLICATION_PORT;
            // Convert the message string into a byte array format (!C# chars are two bytes each!)
            byte[] bBuffer = Encoding.ASCII.GetBytes(sRequest);
            try
            {
                m_RegisterSock.SendTo(bBuffer, m_DirectoryService_IPEndPoint);
            }
            catch
            {   // Handle 'socket not created' or other problems
            }
	        if(false == bRegisterAlreadyDone)
	        {
		        WriteEventLogEntry("Registered with Directory Service");
		        bRegisterAlreadyDone = true;
	        }
        }

        void Start_Registration_Timer()
        {
            m_Registration_Timer = new System.Windows.Forms.Timer();
            m_Registration_Timer.Interval = 10000;	// 10 second interval between auto-registration activity
            m_Registration_Timer.Tick += OnTimedEvent_Registration_Timer;
            m_Registration_Timer.Enabled = true;
        }

        void OnTimedEvent_Registration_Timer(Object myObject, EventArgs myEventArgs)
        {
            SendRegisterMessage();	// Periodic auto-registration
        }
       
        void Start_ApplicationMessage_Timer()
        {
            m_ApplicationMessage_Timer = new System.Windows.Forms.Timer();
            m_ApplicationMessage_Timer.Interval = 50; // 50 milli-second interval for instances of accept and recv activity
            m_ApplicationMessage_Timer.Tick += OnTimedEvent_ApplicationMessage_Timer;
            m_ApplicationMessage_Timer.Enabled = true;
        }

        void OnTimedEvent_ApplicationMessage_Timer(Object myObject, EventArgs myEventArgs)
        {
            Accept();	// Deal with any pending connection requests

            for (int iIndex = 0; iIndex < MAX_CONNECTIONS; iIndex++)	// Deal with any pending messages in the buffer
            {
                if (m_ConnectionArray[iIndex].bInUse == true)
                {
                    Receive(iIndex);	// For each valid listen-socket DoReceive()
                }
            }
        }

        void Accept()
        {
            int iIndex = GetIndexOfUnusedConnection();
            if (iIndex == -1)
            {
                WriteEventLogEntry("Accept refused - Maximum number of Connections reached");
                return;
            }
            try
            {
                m_ConnectionArray[iIndex].m_ConnectSock = m_ListenSock.Accept();
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    WriteEventLogEntry("Accept failed");
                }
                return;
            }

            m_ConnectionArray[iIndex].bInUse = true;
            // Set the newly-created connection socket to Non-Blocking
            m_ConnectionArray[iIndex].m_ConnectSock.Blocking = false;
            IPEndPoint remoteIpEndPoint = m_ConnectionArray[iIndex].m_ConnectSock.RemoteEndPoint as IPEndPoint;
            m_ConnectionArray[iIndex].m_ConnectIPEndPoint = remoteIpEndPoint;
            byte[] bAddress = m_ConnectionArray[iIndex].m_ConnectIPEndPoint.Address.GetAddressBytes();
            string sStr = string.Format("Client {0}.{1}.{2}.{3} Connected",
                    bAddress[0], bAddress[1], bAddress[2], bAddress[3]);
            WriteEventLogEntry(sStr);
        }

        void Receive(int iConnection)
        {		
            byte[] bBuffer = new byte[EventNotificationService_Message_PDU_Size];  // Create an appropriately-sized buffer to Receive into
            int iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount = m_ConnectionArray[iConnection].m_ConnectSock.Receive(bBuffer, 0);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   // check if connection has been lost  
                    if (ex.NativeErrorCode.Equals(10053/*WSAECONNABORTED*/) ||
                        ex.NativeErrorCode.Equals(10054/*WSAECONNRESET*/) ||
                        ex.NativeErrorCode.Equals(10060/*WSAETIMEDOUT*/) ||
                        ex.NativeErrorCode.Equals(10052/*WSAENETRESET*/) ||
                        ex.NativeErrorCode.Equals(10058/*WSAESHUTDOWN*/))
                    {
                        CloseConnection(iConnection);
                    }
                }
            }
            if(EventNotificationService_Message_PDU_Size == iReceiveByteCount)
            {
                // Deserialise the Receive buffer to get the ENS PDU message structure
                EventNotificationService_Message_PDU Message = DeSerialize_ENS_Message(bBuffer);

                string sStr;
                string sStrEvent = ByteArrayTo_NullTerminated_ASCII_String(Message.cEventType);
                string sStrValue = ByteArrayTo_NullTerminated_ASCII_String(Message.cEventValue);
                int iEventTypeList_Index;
		        switch(Message.iMessageType)
		        {
			        case (byte)EventNotificationService_Message_Type.SubscribeToEvent:
                        sStr = string.Format("SubscribeToEvent received: {0}", sStrEvent);
                        WriteEventLogEntry(sStr);
				        AddEventType(Message, iConnection);
				        DisplayEventTypeList();	// Refresh the list of known event types
				        break;
			        case (byte)EventNotificationService_Message_Type.PublishEvent:
                        sStr = string.Format("PublishEvent received: {0} {1}", sStrEvent, sStrValue);
                        WriteEventLogEntry(sStr);
				        iEventTypeList_Index = AddEventType(Message, iConnection);
				        DisplayEventTypeList();	// Refresh the list of known event types
				        PushEventToRegisteredSubscribers(iEventTypeList_Index);
				        break;
			        case (byte)EventNotificationService_Message_Type.UnSubscribeToEvent:
                        sStr = string.Format("UnSubscribeToEvent received: {0}", sStrEvent);
                        WriteEventLogEntry(sStr);
				        RemoveSubscription(Message, iConnection);
				        DisplayEventTypeList();	// Refresh the list of known event types
				        break;
		        }
	        }
        }

        int GetIndexOfEvent(EventNotificationService_Message_PDU Message, bool bCreateNewIfNotExist)
        {
            string sEvent = ByteArrayTo_NullTerminated_ASCII_String(Message.cEventType);
            int iIndex;
	        // Check if the Event Type is already known
	        for(iIndex = 0; iIndex < MAX_EVENT_TYPES; iIndex++)
	        {
                if (true == m_EventTypeArray[iIndex].bInUse_SubscriptionExists || true == m_EventTypeArray[iIndex].bInUse_ValueValid)
		        {
                    if(true == m_EventTypeArray[iIndex].sEventType.Equals(sEvent))
			        {	
                        WriteEventLogEntry("EventType already known");
				        return iIndex;
			        }
		        }
	        }	
	
	        if(false == bCreateNewIfNotExist)
	        {   // Do not create a new entry if the EventType was not found (for UnSubscribe)
                WriteEventLogEntry("EventType not found");
		        return -1; // Event entry was not found
	        }
		
	        // Find an unused m_EventTypeArray entry (for Subscribe, Publish)
	        for(iIndex = 0; iIndex < MAX_EVENT_TYPES; iIndex++)
	        {	
		        if(false == m_EventTypeArray[iIndex].bInUse_SubscriptionExists && false == m_EventTypeArray[iIndex].bInUse_ValueValid)
		        {
                    WriteEventLogEntry("Vacant EventType entry found");
			        // Add the new event type in the vacent list entry
			        m_EventTypeArray[iIndex].sEventType = sEvent; 
			        return iIndex;
		        }
	        }
            WriteEventLogEntry("EventType list FULL !");
	        return -1;
        }

        int AddEventType(EventNotificationService_Message_PDU Message, int iConnection)
        {
	        int iIndex;
	        switch(Message.iMessageType) {
		        case (byte)EventNotificationService_Message_Type.SubscribeToEvent:
			        iIndex = GetIndexOfEvent(Message, true);  // Returns index of event (creates a new event if the EventType does not already exist)
			        m_EventTypeArray[iIndex].bInUse_SubscriptionExists = true;	
			        AddSubscriber_to_SubscriberList(iIndex, iConnection);	// Add subscriber details to subscriber list for event type
			        if(true == m_EventTypeArray[iIndex].bInUse_ValueValid)
			        {   // The event had already been published, so push the current value out to the new subscriber immediately
				        Send_EventUpdate_Message(iIndex, iConnection);
			        }
			        return iIndex;
		        case (byte)EventNotificationService_Message_Type.PublishEvent:
			        iIndex = GetIndexOfEvent(Message, true);  // Returns index of event (creates a new event if the EventType does not already exist)
			        m_EventTypeArray[iIndex].bInUse_ValueValid = true;                   
                    if(0 == Message.cEventValue[0])
			        {   // If EventValue string is blank set it to "0"
				        Message.cEventValue = StringTo_ASCII_ByteArray("0");
			        }
			        m_EventTypeArray[iIndex].sEventValue = ByteArrayTo_NullTerminated_ASCII_String(Message.cEventValue);
			
                    byte[] bAddress = new byte[4];
                    bAddress[0] = Message.SendersAddress.sin_addr.s_b1;
                    bAddress[1] = Message.SendersAddress.sin_addr.s_b2;
                    bAddress[2] = Message.SendersAddress.sin_addr.s_b3;
                    bAddress[3] = Message.SendersAddress.sin_addr.s_b4;
                    IPAddress SenderAddress = new IPAddress(bAddress);
                    m_EventTypeArray[iIndex].PublisherAddress = new IPEndPoint(SenderAddress, Message.SendersAddress.sin_port);
			        return iIndex;
	        }
	        return -1;
        }

        void AddSubscriber_to_SubscriberList(int iIndex, int iConnection)
        {
            SubscriberListItem ItemToAdd = new SubscriberListItem();
            ItemToAdd.iConnectionIndex = iConnection;
    
            // Check to see if the current subscriber is already in the list
            LinkedListNode<SubscriberListItem> FoundItem = m_EventTypeArray[iIndex].SubscriberList.Find(ItemToAdd);
            if(null != FoundItem)
		    {
			    return;   // Subscriber already in the list for the particular EventType
		    }
            m_EventTypeArray[iIndex].SubscriberList.AddFirst(ItemToAdd);
        }

        void PushEventToRegisteredSubscribers(int iEventTypeList_Index)
        {
	        // Iterate through the list, send an Event update message to each subscriber in the list
            foreach (var val in m_EventTypeArray[iEventTypeList_Index].SubscriberList)
            {
                Send_EventUpdate_Message(iEventTypeList_Index, val.iConnectionIndex);
	        }
        }

        void Send_EventUpdate_Message(int iEventTypeList_Index, int iConnectionIndex)
        {
	        EventNotificationService_Message_PDU ENS_Message;
	        ENS_Message.iMessageType = (byte)EventNotificationService_Message_Type.EventUpdate;
	        ENS_Message.cEventType = StringTo_ASCII_ByteArray(m_EventTypeArray[iEventTypeList_Index].sEventType);
            ENS_Message.cEventValue = StringTo_ASCII_ByteArray(m_EventTypeArray[iEventTypeList_Index].sEventValue);
            ENS_Message.SendersAddress = m_LocalAddress_FormattedAs_SockAddr_In;
            SendMessageToClient(ENS_Message, iConnectionIndex);
        }

        void SendMessageToClient(EventNotificationService_Message_PDU ENS_Message, int iConnectionIndex)
        {
            byte[] bBuffer = Serialize_ENS_PDU(ENS_Message);
            int iBytesSent = 0;
            try
            {
                iBytesSent = m_ConnectionArray[iConnectionIndex].m_ConnectSock.Send(bBuffer);
                if (0 < iBytesSent)
                {
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   // check if connection has been lost  
                    if (ex.NativeErrorCode.Equals(10053/*WSAECONNABORTED*/) ||
                        ex.NativeErrorCode.Equals(10054/*WSAECONNRESET*/) ||
                        ex.NativeErrorCode.Equals(10060/*WSAETIMEDOUT*/) ||
                        ex.NativeErrorCode.Equals(10052/*WSAENETRESET*/) ||
                        ex.NativeErrorCode.Equals(10058/*WSAESHUTDOWN*/))
                    {
                        CloseConnection(iConnectionIndex);
                    }
                }
            }
	        if(0 < iBytesSent)
	        {
                WriteEventLogEntry("Message sent to ENS client");
	        }
        }

        void RemoveSubscription(EventNotificationService_Message_PDU Message, int iConnection)
        {
	        // Find the relevent entry in linked list
	        int iIndex;
	        iIndex = GetIndexOfEvent(Message, false);  // Returns index of event (does not create a new event if the EventType does not already exist)
	        if(-1 == iIndex)
	        {
		        return;	// EventType entry not found in array
	        }
	        RemoveSubscriber_from_SubscriberList(iIndex, iConnection);	// Remove subscriber details from subscriber list for event type
	
	        if(0 == m_EventTypeArray[iIndex].SubscriberList.Count)
	        {	// If subscriber list now emply, set subscribed flag  to false
		        m_EventTypeArray[iIndex].bInUse_SubscriptionExists = false;
	        }

	        // If both flags are now false (the bInUse_SubscriptionExists flag and the bInUse_ValueValid flag) the EventType
	        // is essentially removed, as its array entry will be re-used when a new EventType is created
        }

        void RemoveSubscriber_from_SubscriberList(int iIndex, int iConnection)
        {
            SubscriberListItem ItemToRemove = new SubscriberListItem();
            ItemToRemove.iConnectionIndex = iConnection;

            // Check to see if the current subscriber is in the list
            LinkedListNode<SubscriberListItem> FoundItem = m_EventTypeArray[iIndex].SubscriberList.Find(ItemToRemove);
            if (null != FoundItem)
            {
                m_EventTypeArray[iIndex].SubscriberList.Remove(FoundItem);
	        }
        }

        int GetIndexOfUnusedConnection()
        {
	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
		        if(m_ConnectionArray[i].bInUse == false)
		        {
			        return i;
		        }
	        }
	        return -1;	// Signal that all connections are in use
        }

        private void Done_button_Click(object sender, EventArgs e)
        {   // Close the socket and exit the application
            CloseAndExit();
        }

        private void CloseAndExit()
        {   // Close all sockets and exit the application
            for (int i = 0; i < MAX_CONNECTIONS; i++)	// Close active connections
            {
                if (m_ConnectionArray[i].bInUse == true)
                {
                    CloseConnection(i);
                }
            }
            try
            {
                m_RegisterSock.Close();
            }
            catch
            {
            }
            try
            {
                m_ListenSock.Close();
            }
            catch
            {
            }
            System.Environment.Exit(0);
        }

        void CloseConnection(int iConnection)
        {
            try
            {
                m_ConnectionArray[iConnection].m_ConnectSock.Shutdown(SocketShutdown.Both);	// Close the connection
                m_ConnectionArray[iConnection].m_ConnectSock.Close();			// Close the socket
            }
            catch
            {
            }
            m_ConnectionArray[iConnection].bInUse = false;

            byte[] bAddress = m_ConnectionArray[iConnection].m_ConnectIPEndPoint.Address.GetAddressBytes();
            string sStr = string.Format("Client {0}.{1}.{2}.{3} Disconnected",
                    bAddress[0], bAddress[1], bAddress[2], bAddress[3]);
                    WriteEventLogEntry(sStr);
        }

        public EventNotificationService_Message_PDU DeSerialize_ENS_Message(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a EventNotificationService_Message_PDU structure (Fields have fixed length, but non-fixed data length)
            EventNotificationService_Message_PDU ENS_PDU = new EventNotificationService_Message_PDU();
            ENS_PDU.iMessageType = bRecdByteArray[0];
            ENS_PDU.cEventType = new byte[MAX_EVENT_TYPE_NAME_LENGTH];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 1, ENS_PDU.cEventType/*dest*/, 0, MAX_EVENT_TYPE_NAME_LENGTH/*count*/);
            ENS_PDU.cEventValue = new byte[MAX_EVENT_VALUE_LENGTH];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 51, ENS_PDU.cEventValue/*dest*/, 0, MAX_EVENT_VALUE_LENGTH/*count*/);
            ENS_PDU.SendersAddress.sin_family = bRecdByteArray[104];
            // DeSerialisation of port number
            // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format: high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received by clients, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            ENS_PDU.SendersAddress.sin_port = (Int16)((bRecdByteArray[106] * 256) + bRecdByteArray[107]);
            ENS_PDU.SendersAddress.sin_addr.s_b1 = bRecdByteArray[108];
            ENS_PDU.SendersAddress.sin_addr.s_b2 = bRecdByteArray[109];
            ENS_PDU.SendersAddress.sin_addr.s_b3 = bRecdByteArray[110];
            ENS_PDU.SendersAddress.sin_addr.s_b4 = bRecdByteArray[111];
            return ENS_PDU;
        }

        private byte[] Serialize_ENS_PDU(EventNotificationService_Message_PDU ENS_PDU)
        {   // Convert an EventNotificationService_Message_PDU into a byte array
            byte[] bByteArray = new byte[EventNotificationService_Message_PDU_Size];        // Create a byte array to hold the serialised form of the ENS_PDU
            bByteArray[0] = (byte)ENS_PDU.iMessageType;
            Buffer.BlockCopy(ENS_PDU.cEventType/*src*/, 0, bByteArray/*dest*/, 1, ENS_PDU.cEventType.Length);
            Buffer.BlockCopy(ENS_PDU.cEventValue/*src*/, 0, bByteArray/*dest*/, 51, ENS_PDU.cEventValue.Length);
            bByteArray[104] = (byte)ENS_PDU.SendersAddress.sin_family;
            // Serialisation of port number
            // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format: high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received by clients, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            bByteArray[106] = (byte)(ENS_PDU.SendersAddress.sin_port / 256); // Create the MSB byte of the port value
            bByteArray[107] = (byte)(ENS_PDU.SendersAddress.sin_port % 256); // Create the LSB byte of the port value
            bByteArray[108] = ENS_PDU.SendersAddress.sin_addr.s_b1;
            bByteArray[109] = ENS_PDU.SendersAddress.sin_addr.s_b2;
            bByteArray[110] = ENS_PDU.SendersAddress.sin_addr.s_b3;
            bByteArray[111] = ENS_PDU.SendersAddress.sin_addr.s_b4;
            return bByteArray;
        }

        private void WriteEventLogEntry(string sStr)
        {
            Event_Log_listBox.Items.Add(sStr);       
        }
        
        private void ClearEventLog_button_Click(object sender, EventArgs e)
        {
            Event_Log_listBox.Items.Clear();
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static string ByteArrayTo_NullTerminated_ASCII_String(byte[] bArray) // Returns string terminated at the null terminator
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            string sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            string sFinal = sStr.Substring(0, iASCIIz_length);
            return sFinal;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }
    }
}