﻿/*	File			CaseStudy_NotServ_Consume_Client_CS.cs
	Purpose			Notification Service Client  (Publisher)
	Author			Richard Anthony	(R.J.Anthony@gre.ac.uk)
	Date			September 2014
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;

namespace CaseStudy_NotServ_Consume_Client_CSharp
{
    public partial class CaseStudy_NotServ_Consume_Client_CSharp : Form
    {   // Define constants 
        public const string EVENT_NOTIFICATION_SERVER_NAME = "Event_NS";	// Event Notification Server
        public const int MAX_EVENT_TYPE_NAME_LENGTH = 50;
        public const int MAX_EVENT_VALUE_LENGTH = 50;
        public const int DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH = 50;

        public const uint From_DIRECTORY_SERVICE_PORT = 8002;	// Fixed, defined by Directory Service
        public const uint ENS_APPLICATION_PORT = 8003;          // TCP, ENS_Server Binds to this port
        public const uint To_DIRECTORY_SERVICE_PORT = 8008;	    // Fixed, defined by Directory Service

        // Define communication types
        [StructLayout(LayoutKind.Sequential)]
        public struct DirectoryServiceRequest_PDU
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH)]
            public byte[] bBuffer;
        };

        [StructLayout(LayoutKind.Sequential)]
        public struct DirectoryServiceReply_PDU                 // Fixed, defined by Directory Service
        {   // IP address followed by port
            public byte a1;
            public byte a2;
            public byte a3;
            public byte a4;
            public uint iPort;
        };

        enum EventNotificationService_Message_Type { SubscribeToEvent, UnSubscribeToEvent, PublishEvent, EventUpdate };

        // SOCKADDR_IN not defined in c# - define it and its constituent sub-structures
        // Needed because the ENS passes the message sender's address in this format
        [StructLayout(LayoutKind.Sequential)]
        public unsafe struct SockAddr
        {
            public ushort sa_family;
            public fixed byte sa_data[14]; // Note: use byte instead of char because sizeof(char) == 2 in C#
        }

        [StructLayout(LayoutKind.Sequential)]
        public unsafe struct SockAddr_In
        {
            public Int16 sin_family;
            public Int16 sin_port;
            public In_Addr sin_addr;
            public fixed byte sin_zero[8];
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct In_Addr
        {
            public byte s_b1, s_b2, s_b3, s_b4;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct EventNotificationService_Message_PDU
        {
            public byte iMessageType;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_EVENT_TYPE_NAME_LENGTH)]
            public byte[] cEventType;       //  (c++)  char[] cEventType[MAX_EVENT_TYPE_NAME_LENGTH];
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_EVENT_VALUE_LENGTH)]
            public byte[] cEventValue;      //  (c++)  char[] cEventValue[MAX_EVENT_VALUE_LENGTH];
            // In actual transmission, a 3-byte pad is placed here so that the SockAddr_In begins on a 4-byte boundary
            public SockAddr_In SendersAddress;
        };

        public const Int32 EventNotificationService_Message_PDU_Size = 120;

        // Define member variables
        Socket m_ResolveSock;
        Socket m_ApplicationSock;
        IPEndPoint m_DirectoryService_IPEndPoint;
        IPEndPoint m_localIPEndPoint;
        IPEndPoint m_EventNotificationServerAddress;
        System.Windows.Forms.Timer m_DirectoryServiceReply_Timer;
        System.Windows.Forms.Timer m_ApplicationMessage_Timer;
        bool m_bConnected;
        EventNotificationService_Message_PDU m_ENS_PDU;
        bool m_bSubscribedToTemperatureEvent;
        bool m_bSubscribedToHumidityEvent;

        public CaseStudy_NotServ_Consume_Client_CSharp()
        {
            InitializeComponent();
            m_bConnected = false;
            m_bSubscribedToTemperatureEvent = false;
            m_bSubscribedToHumidityEvent = false;
            TemperatureEvent_button.Enabled = false;
            HumidityEvent_button.Enabled = false;

            if (true == CreateResolveSocket())
            {
                bool bResult = BindResolveSocket();
                if (true == bResult)
                {
                    ContactDirectoryService(EVENT_NOTIFICATION_SERVER_NAME);
                }
            }
        }

        private void ContactDirectoryService(string NameToResolve)
        {    // Initialise timer to wait for directory service reply (uses non-blocking socket)
            m_DirectoryServiceReply_Timer = new System.Windows.Forms.Timer();
            m_DirectoryServiceReply_Timer.Interval = 250; // 250 millisecond period
            m_DirectoryServiceReply_Timer.Tick += new EventHandler(On_DirectoryServiceReply_Timer);
            m_DirectoryServiceReply_Timer.Enabled = true;
            SendResolveMessage();
        }

        bool CreateResolveSocket()
        {   // UDP, non-blocking, broadcast
            try
            {
                m_ResolveSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_ResolveSock.Blocking = false;
                m_ResolveSock.EnableBroadcast = true;
                System.Net.IPAddress BroadcastIPAddress = System.Net.IPAddress.Broadcast;
                m_DirectoryService_IPEndPoint = new System.Net.IPEndPoint(BroadcastIPAddress, System.Convert.ToInt16(To_DIRECTORY_SERVICE_PORT));
                Event_Log_listBox.Items.Add("Resolve Socket created");
                return true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                Event_Log_listBox.Items.Add("Create Socket failed");
                return false;
            }
            catch   // Catch any other exceptions (e.g. arising from creating the Endpoint)
            {
                Event_Log_listBox.Items.Add("Invalid IP address or Port number");
                return false;
            }
        }

        void SendResolveMessage() // Broadcast message to Directory Server
        {
            String sRequest = "RESOLVE:" + EVENT_NOTIFICATION_SERVER_NAME;
            // Convert the message string into a byte array format (!C# chars are two bytes each!)
            byte[] bBuffer = Encoding.ASCII.GetBytes(sRequest);
            m_ResolveSock.SendTo(bBuffer, m_DirectoryService_IPEndPoint);
        }

        public bool BindResolveSocket()
        {
            // Create an Endpoint that will cause the listening activity to apply to all the local node's interfaces
            m_localIPEndPoint = new System.Net.IPEndPoint(IPAddress.Any, System.Convert.ToInt16(From_DIRECTORY_SERVICE_PORT));
            try
            {   // Bind to the local IP Address and selected port
                m_ResolveSock.Bind(m_localIPEndPoint);
                return true;
            }
            catch   // Catch any exceptions
            {
                Event_Log_listBox.Items.Add("bind failed");
                return false;
            }
        }

        private void On_DirectoryServiceReply_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                Receive_DirectoryServiceReply();
            }
            catch // Catch any errors
            {
            }
        }

        private void Receive_DirectoryServiceReply()
        {
            EndPoint localEndPoint = (EndPoint)m_localIPEndPoint;  // New EndPoint structure to hold the address of the sender 
            byte[] bBuffer = new byte[DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH];
            int iReceiveByteCount;
            iReceiveByteCount = m_ResolveSock.ReceiveFrom(bBuffer, ref localEndPoint);
            if (0 < iReceiveByteCount)
            {   // Stop the timer now the Directory Service reply has been received
                m_DirectoryServiceReply_Timer.Enabled = false;
                // Close the resolve socket, so the Directory Service port is released and thus other Directory Service clients can run 
                // on the same computer (this is very useful for development and testing purposes)
                m_ResolveSock.Close();
                // DeSerialise the message received                
                DirectoryServiceReply_PDU DirectoryServiceReply_message = DeSerialize_DirectoryServiceReply(bBuffer);
                if (DirectoryServiceReply_message.iPort != 0)	// Port of 0 means App Service was not known
                {   // Use the resolved ENS address information to create a TCP connection with the ENS server
                    if (false == m_bConnected)
                    {
                        CreateApplicationSocket();
                        SetUpSendAddress_to_EventNS(DirectoryServiceReply_message);
                        ConnectToENS();
                    }
                }
            }
        }

        public DirectoryServiceReply_PDU DeSerialize_DirectoryServiceReply(byte[] received)
        {   // Converts a byte array to a DirectoryServiceReply_PDU structure (all fields fixed data length)
            IntPtr ip = Marshal.AllocHGlobal(received.Length);  // Allocate unmanaged memory (working storage)
            Marshal.Copy(received, 0, ip, received.Length);     //Copy the byte array to the working storage
            // Marshal the unmanaged memory contents to the DirectoryServiceReply_PDU structure
            DirectoryServiceReply_PDU structure = (DirectoryServiceReply_PDU)Marshal.PtrToStructure(ip, typeof(DirectoryServiceReply_PDU));
            Marshal.FreeHGlobal(ip);                            // Free working storage
            return structure;
        }

        public EventNotificationService_Message_PDU DeSerialize_ENS_Message(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a EventNotificationService_Message_PDU structure (Fields have fixed length, but non-fixed data length)
            EventNotificationService_Message_PDU ENS_PDU = new EventNotificationService_Message_PDU();
            ENS_PDU.iMessageType = bRecdByteArray[0];
            ENS_PDU.cEventType = new byte[MAX_EVENT_TYPE_NAME_LENGTH];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 1, ENS_PDU.cEventType/*dest*/, 0, MAX_EVENT_TYPE_NAME_LENGTH/*count*/);
            ENS_PDU.cEventValue = new byte[MAX_EVENT_VALUE_LENGTH];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 51, ENS_PDU.cEventValue/*dest*/, 0, MAX_EVENT_VALUE_LENGTH/*count*/);
            ENS_PDU.SendersAddress.sin_family = bRecdByteArray[104];
            // DeSerialisation of port number
            // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format: high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received by clients, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            ENS_PDU.SendersAddress.sin_port = (Int16)((bRecdByteArray[106] * 256) + bRecdByteArray[107]);
            ENS_PDU.SendersAddress.sin_addr.s_b1 = bRecdByteArray[108];
            ENS_PDU.SendersAddress.sin_addr.s_b2 = bRecdByteArray[109];
            ENS_PDU.SendersAddress.sin_addr.s_b3 = bRecdByteArray[110];
            ENS_PDU.SendersAddress.sin_addr.s_b4 = bRecdByteArray[111];
            return ENS_PDU;
        }

        bool CreateApplicationSocket()
        {   // TDP, blocking initially
            try
            {
                m_ApplicationSock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ApplicationSock.Blocking = true;  // Blocking mode is used initially, for Connect()
                Event_Log_listBox.Items.Add("Application Socket created");
                return true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                return false;
            }
            catch   // Catch any other exceptions (e.g. arising from creating the Endpoint)
            {
                return false;
            }
        }

        void SetUpSendAddress_to_EventNS(DirectoryServiceReply_PDU DirectoryServiceReply_message)
        {	// The ENS server address and port is retrieved from the Directory Service
            byte[] bAddress = new byte[4];
            bAddress[0] = DirectoryServiceReply_message.a1;
            bAddress[1] = DirectoryServiceReply_message.a2;
            bAddress[2] = DirectoryServiceReply_message.a3;
            bAddress[3] = DirectoryServiceReply_message.a4;
            IPAddress ENS_Address = new IPAddress(bAddress);
            m_EventNotificationServerAddress = new System.Net.IPEndPoint(ENS_Address, System.Convert.ToInt16(DirectoryServiceReply_message.iPort));
        }

        bool ConnectToENS()
        {   // Initiate TCP connection with the ENS server
            try
            {
                m_ApplicationSock.Connect(m_EventNotificationServerAddress);
                m_ApplicationSock.Blocking = false;  // set Non-Blocking mode for Receive()
            }
            catch   // Catch any exceptions
            {
                m_ApplicationSock.Close();
                return false;
            }

            Event_Log_listBox.Items.Add("Connected to ENS");
            TemperatureEvent_button.Enabled = true;
            HumidityEvent_button.Enabled = true;
            m_bConnected = true;

            // Initialise timer to wait for ENS service reply (uses non-blocking socket)
            m_ApplicationMessage_Timer = new System.Windows.Forms.Timer();
            m_ApplicationMessage_Timer.Interval = 50;                     // 250 millisecond period
            m_ApplicationMessage_Timer.Tick += On_ApplicationMessage_Timer;
            m_ApplicationMessage_Timer.Enabled = true;
            return true;
        }

        private void On_ApplicationMessage_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                Receive_ENS_Reply();
            }
            catch // Silently catch any errors
            {
            }
        }

        private void Receive_ENS_Reply()
        {
            byte[] bBuffer = new byte[EventNotificationService_Message_PDU_Size];  // Create an appropriately-sized buffer to Receive into
            int iReceiveByteCount;
            iReceiveByteCount = m_ApplicationSock.Receive(bBuffer, 0);
            if (iReceiveByteCount == EventNotificationService_Message_PDU_Size)
            {
                m_ENS_PDU = DeSerialize_ENS_Message(bBuffer); // Deserialise the Receive buffer to get the ENS PDU message structure 
                // Additional deserialisation actions needed to extract ASCIIz strings from the byte array fieds
                string sEventType = Encoding.ASCII.GetString(m_ENS_PDU.cEventType);
                string sEventValue = Encoding.UTF8.GetString(m_ENS_PDU.cEventValue);
                string sEventTypeTruncated = StringExt.Truncate_To_ASCIIz(sEventType); // See extension class StringExt (at end of code listing)
                string sEventValueTruncated = StringExt.Truncate_To_ASCIIz(sEventValue);
                string sStr = "Received: ";
                sStr += sEventTypeTruncated;
                sStr += " Value: ";
                sStr += sEventValueTruncated;
                Event_Log_listBox.Items.Add(sStr);
                if ("TemperatureChange" == sEventTypeTruncated)
                {
                    Temperature_textBox.Text = sEventValue;
                }
                if ("HumidityChange" == sEventTypeTruncated)
                {
                    Humidity_textBox.Text = sEventValue;
                }
            }
        }

        public SockAddr_In GetLocalIPAddress()
        {
            string szHost = Dns.GetHostName();
            SockAddr_In SockAddr_In_LocalAddress = new SockAddr_In();
            SockAddr_In_LocalAddress.sin_addr.s_b1 = 127;   // Default is local loopback address
            SockAddr_In_LocalAddress.sin_addr.s_b1 = 0;
            SockAddr_In_LocalAddress.sin_addr.s_b1 = 0;
            SockAddr_In_LocalAddress.sin_addr.s_b1 = 1;

            IPHostEntry IPHost = Dns.GetHostEntry(Dns.GetHostName());
            foreach (IPAddress IP in IPHost.AddressList)
            {
                if (IP.AddressFamily == AddressFamily.InterNetwork) // Match only the IPv4 address
                {   // Get first address from list
                    byte[] bAddress = IP.GetAddressBytes();
                    SockAddr_In_LocalAddress.sin_addr.s_b1 = bAddress[0];
                    SockAddr_In_LocalAddress.sin_addr.s_b2 = bAddress[1];
                    SockAddr_In_LocalAddress.sin_addr.s_b3 = bAddress[2];
                    SockAddr_In_LocalAddress.sin_addr.s_b4 = bAddress[3];
                    SockAddr_In_LocalAddress.sin_port = 0;  // Port value does not currently discriminate between clients at same address
                    break;
                }
            }
            return SockAddr_In_LocalAddress;
        }

        private byte[] Serialize_ENS_PDU(EventNotificationService_Message_PDU ENS_PDU)
        {   // Convert an EventNotificationService_Message_PDU into a byte array
            byte[] bByteArray = new byte[EventNotificationService_Message_PDU_Size];        // Create a byte array to hold the serialised form of the ENS_PDU
            bByteArray[0] = (byte)ENS_PDU.iMessageType;
            Buffer.BlockCopy(ENS_PDU.cEventType/*src*/, 0, bByteArray/*dest*/, 1, ENS_PDU.cEventType.Length);
            Buffer.BlockCopy(ENS_PDU.cEventValue/*src*/, 0, bByteArray/*dest*/, 51, ENS_PDU.cEventValue.Length);
            bByteArray[104] = (byte)ENS_PDU.SendersAddress.sin_family;
            // Serialisation of port number
            // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format: high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received by clients, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            bByteArray[106] = (byte)(ENS_PDU.SendersAddress.sin_port / 256); // Create the MSB byte of the port value
            bByteArray[107] = (byte)(ENS_PDU.SendersAddress.sin_port % 256); // Create the LSB byte of the port value
            bByteArray[108] = ENS_PDU.SendersAddress.sin_addr.s_b1;
            bByteArray[109] = ENS_PDU.SendersAddress.sin_addr.s_b2;
            bByteArray[110] = ENS_PDU.SendersAddress.sin_addr.s_b3;
            bByteArray[111] = ENS_PDU.SendersAddress.sin_addr.s_b4;
            return bByteArray;
        }

        void SendMessageToAppServer(EventNotificationService_Message_PDU ENS_Message)
        {
            byte[] bBuffer = Serialize_ENS_PDU(ENS_Message);
            int iBytesSent;
            if (true == m_bConnected)
            {
                try
                {
                    iBytesSent = m_ApplicationSock.Send(bBuffer);
                    if (0 < iBytesSent)
                    {
                    }
                }
                catch // Catch any errors
                {
                    Event_Log_listBox.Items.Add("Send error");
                }
            }
        }

        private void Done_button_Click(object sender, EventArgs e)
        {   // Close the socket and exit the application
            try
            {
                m_ResolveSock.Close();
                m_ApplicationSock.Close();
            }
            catch
            {
            }
            this.Close();
        }

        private void TemperatureEvent_button_Click(object sender, EventArgs e)
        {
            EventNotificationService_Message_PDU ENS_Message = new EventNotificationService_Message_PDU();
            ENS_Message.SendersAddress = GetLocalIPAddress();           // Port value does not currently discriminate between clients at same address
            String sStr = "TemperatureChange";
            ENS_Message.cEventType = Encoding.ASCII.GetBytes(sStr);     // Convert message string into byte array format (!C# chars are two bytes each!)
            sStr = "";
            ENS_Message.cEventValue = Encoding.ASCII.GetBytes(sStr);    // Convert message string into byte array format (!C# chars are two bytes each!)
 
	        if(false == m_bSubscribedToTemperatureEvent)
	        {
		        m_bSubscribedToTemperatureEvent = true;
		        ENS_Message.iMessageType = (byte)EventNotificationService_Message_Type.SubscribeToEvent;
 
		        TemperatureEvent_button.Text = "UnSubscribe to TemperatureChange events";
	        }
	        else
	        {
		        m_bSubscribedToTemperatureEvent = false;
                ENS_Message.iMessageType = (byte)EventNotificationService_Message_Type.UnSubscribeToEvent;
                TemperatureEvent_button.Text = "Subscribe to TemperatureChange events";
	        }
            Temperature_textBox.Text = "";
            SendMessageToAppServer(ENS_Message);
            Event_Log_listBox.Items.Add("Message sent to ENS Service");
        }

        private void HumidityEvent_button_Click(object sender, EventArgs e)
        {
            EventNotificationService_Message_PDU ENS_Message = new EventNotificationService_Message_PDU();
            ENS_Message.SendersAddress = GetLocalIPAddress();           // Port value does not currently discriminate between clients at same address
            ENS_Message.iMessageType = (byte)EventNotificationService_Message_Type.SubscribeToEvent;
            String sStr = "HumidityChange";
            ENS_Message.cEventType = Encoding.ASCII.GetBytes(sStr);     // Convert message string into byte array format (!C# chars are two bytes each!)
            sStr = "";
            ENS_Message.cEventValue = Encoding.ASCII.GetBytes(sStr);    // Convert message string into byte array format (!C# chars are two bytes each!)
            
            if (false == m_bSubscribedToHumidityEvent)
            {
                m_bSubscribedToHumidityEvent = true;
                ENS_Message.iMessageType = (byte)EventNotificationService_Message_Type.SubscribeToEvent;

                HumidityEvent_button.Text = "UnSubscribe to HumidityChange events";
            }
            else
            {
                m_bSubscribedToHumidityEvent = false;
                ENS_Message.iMessageType = (byte)EventNotificationService_Message_Type.UnSubscribeToEvent;
                HumidityEvent_button.Text = "Subscribe to HumidityChange events";
            }
            Humidity_textBox.Text = "";
            SendMessageToAppServer(ENS_Message);
            Event_Log_listBox.Items.Add("Message sent to ENS Service");
         }

        private void ClearEventLog_button_Click(object sender, EventArgs e)
        {
            Event_Log_listBox.Items.Clear();
        }
    }

    public static class StringExt
    {   // Extension class needed for deserialisation of ASCIIz (null terminated strings) held as byte arrays
        public static string Truncate_To_ASCIIz(this string value)
        {
            if (string.IsNullOrEmpty(value))
            {
                return value;
            }
            int iNullTerminatorPosition = value.IndexOf('\0'/*char*/, 0/*start index*/);
            if (-1 == iNullTerminatorPosition)
            {
                return value;
            }
            return value.Substring(0, iNullTerminatorPosition);
        }
    }
}