﻿/*	File			TimeServClnt_Console_AppSide_uses_lib.cs
	Purpose			Network Time Protocol (NTP) Client console application
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			February 2015
*/

using System;
using System.Net;
using System.Net.Sockets;

// Include the namespace of the library that contains the NTP_Client class
// Note that it is also necessary to include the libray in the list of references for this project
// The reference path may need to be ammended if the library path / location is changed
// application project change.
using TimeServiceClient_Library;
// IMPORTANT NOTE: if this project does not build, check that the library .Dll file 
// is included in the list of references for this project. The path of the .Dll file must be 
// correct, see the 'References -> Add Reference' option, and use the browse feature within.

namespace TimeServiceClient
{
    class TimeServClnt_Console_AppSide_uses_lib
    {
        NTP_Client m_NTP_Client;
        Int32 m_iNumRequestsSent;
        Int32 m_iNumResponsesReceived;

        static void Main(string[] args)
        {
            TimeServClnt_Console_AppSide_uses_lib TimeServClnt = new TimeServClnt_Console_AppSide_uses_lib();
            TimeServClnt.InitialiseLibrary();
             
            Console.WriteLine("Time Service Client (Console Application - uses library)");
            Console.WriteLine("--------------------------------------------------------");

            string sDomain = TimeServClnt.Parse_CommandLineArguments(args);
	        Console.WriteLine("Time Service Domain: {0}", sDomain);
            TimeServClnt.SetUp_TimeService_AddressStruct(sDomain);
	        Console.WriteLine("Ask Time Service for timestamp 5 times, at 5-second intervals");
            TimeServClnt.m_NTP_Client.Set_ClientStarted_Flag(true);

            for (int iLoop = 0; iLoop < 5; iLoop++)
	        {
                TimeServClnt.Get_NTP_Timestamp();
		        System.Threading.Thread.Sleep(5000); // 5 second delay between loop iterations
	        }
            TimeServClnt.Quit(0);
        }

        string Parse_CommandLineArguments(string[] args)
        {
            string sProgramName = Environment.GetCommandLineArgs()[0];  // Not the same as args[0]
            if (1 != args.Length)
            {
                // User did not provide the required single option parameter
                // Note, in C++ argv[0] is the program name, argv[1] is the first supplied parameter
                // However in C# args[0] is the first supplied parameter; so need a different technique to get the program name
		        DisplayUsage(sProgramName);
		        Console.WriteLine("!!! Command Line Arguments invalid !!!   QUITTING");
		        Quit(-1);
            }

	        int iOption = Convert.ToInt32(args[0]);  // Resolves to 0 if characters present, therefore options start at 1
	        if(1 > iOption || 12 < iOption)
	        {   // Option value was out of range
		        DisplayUsage(sProgramName);
		        Console.WriteLine("!!! Command Line Arguments invalid !!!   QUITTING");
		        Quit(-2);
	        }
	
	        switch(iOption) {
		        case 1: 
			        return "time.nist.gov";
		        case 2: 
			        return "nist1-ny.ustiming.org";
		        case 3: 
			        return "nist1-nj.ustiming.org";
		        case 4: 
			        return "nist1-atl.ustiming.org";
		        case 5: 
			        return "nist1-chi.ustiming.org";
		        case 6: 
			        return "nist1-lnk.binary.net";
		        case 7: 
			        return "time-a.timefreq.bldrdoc.gov";
		        case 8: 
			        return "ntp-nist.ldsbc.edu";
		        case 9: 
			        return "nist1-lv.ustiming.org";
		        case 10: 
			        return "nist1-la.ustiming.org";
		        case 11: 
			        return "time.windows.com";
		        case 12: 
			        return "wolfnisttime.com";	
	        }
	        return "time.nist.gov";
        }

        void InitialiseLibrary()
        {
            bool bSocketOpenSuccess = false;
            m_NTP_Client = new NTP_Client(ref bSocketOpenSuccess);
            if (false == bSocketOpenSuccess)
	        {   // Library not initialised - could not open socket, or could not set socket to non-blocking mode
		        Console.WriteLine("!!! Failed to initialise Time Service Client library !!!   QUITTING");
		        Quit(-4);
	        }
        }

        void Get_NTP_Timestamp()
        {
            NTP_Client.NTP_Timestamp_Data NTP_Timestamp = new NTP_Client.NTP_Timestamp_Data();
            NTP_Client.NTP_Client_ResultCode Result = m_NTP_Client.Get_NTP_Timestamp(ref NTP_Timestamp);
            switch (Result)
            {
                case NTP_Client.NTP_Client_ResultCode.NTP_Success:
                    m_iNumRequestsSent++;
                    m_iNumResponsesReceived++;

                    Console.WriteLine("Unix Time (seconds sine 1970):{0}", NTP_Timestamp.lUnixTime);
                    Console.WriteLine("Current time   Hour:{0:00} Minute:{1:00} Second:{2:00}", 
                                    NTP_Timestamp.iHour, NTP_Timestamp.iMinute, NTP_Timestamp.iSecond);
                     // NTP does not indicate information specifcally related to local time zones such as BST or other daylight savings time offsets
                    Console.WriteLine("Number of NTP requests sent:{0}  Number of NTP responses received:{1}",
                                                            m_iNumRequestsSent, m_iNumResponsesReceived); 
                    break;
                case NTP_Client.NTP_Client_ResultCode.NTP_ServerAddressNotSet:
                    Console.WriteLine("!!! Time Service Client Library not initialised with Time server Address !!!");
                    break;
                case NTP_Client.NTP_Client_ResultCode.NTP_SendFailed:
                    Console.WriteLine("!!! NTP Send Failed in Time Service Client Library !!!");
                    break;
                case NTP_Client.NTP_Client_ResultCode.NTP_ReceiveFailed:
                    Console.WriteLine("!!! NTP Receive Failed in Time Service Client Library !!!");
                    m_iNumRequestsSent++;
                    Console.WriteLine("Number of NTP requests sent:{0}  Number of NTP responses received:{1}",
                                                            m_iNumRequestsSent, m_iNumResponsesReceived); 
                    break;
            }
        }

        void SetUp_TimeService_AddressStruct(string sURL)
        {
            string sIP_Address1 = "";
            string sIP_Address2 = "";
            string sIP_Address3 = "";
            string sIP_Address4 = "";
            string sPort = "";
            
            m_NTP_Client.SetUp_TimeService_AddressStruct(sURL, ref sIP_Address1, 
                ref sIP_Address2, ref sIP_Address3, ref sIP_Address4, ref sPort);

            Console.WriteLine("Time Service IP address (resolved by the library): {0}.{1}.{2}.{3}:{4}",
                                        sIP_Address1, sIP_Address2, sIP_Address3, sIP_Address4, sPort);
        }
             
        void DisplayUsage(string sProgramName)
        {
	        Console.WriteLine("Usage  {0}  Time-Service-Domain-Name-Option (see below)", sProgramName);	
	        Console.WriteLine("Select from the following Time Service Domain Names (enter the numerical index on the command line):");	
	        Console.WriteLine(" 1   time.nist.gov                (Location: NIST round robin load equalisation");
	        Console.WriteLine(" 2   nist1-ny.ustiming.org        (Location: New York City, NY");	
	        Console.WriteLine(" 3   nist1-nj.ustiming.org        (Location: Bridgewater, NJ");	
	        Console.WriteLine(" 4   nist1-atl.ustiming.org       (Location: Atlanta, Georgia");	
	        Console.WriteLine(" 5   nist1-chi.ustiming.org       (Location: Chicago, Illinois");	
	        Console.WriteLine(" 6   nist1-lnk.binary.net         (Location: Lincoln, Nebraska" );	
	        Console.WriteLine(" 7   time-a.timefreq.bldrdoc.gov  (Location: NIST, Boulder, Colorado");	
	        Console.WriteLine(" 8   ntp-nist.ldsbc.edu           (Location: LDSBC, Salt Lake City, Utah");	
	        Console.WriteLine(" 9   nist1-lv.ustiming.org        (Location: Las Vegas, Nevada");	  
	        Console.WriteLine("10   nist1-la.ustiming.org        (Location: Los Angeles, California");	
	        Console.WriteLine("11   time.windows.com             (Location: Windows Time service");	
	        Console.WriteLine("12   wolfnisttime.com             (Location: Birmingham, Alabama");	
        }

        void Quit(Int32 iExitCode)
        {
            try
            {
                m_NTP_Client.CloseSocket();
            }
            catch
            {   // Necessary because Socket might not yet be created
            }
            Console.WriteLine("Press any key to quit");
	        Console.ReadKey();
	        System.Environment.Exit(iExitCode);
        }
    }
}
