﻿/*	File			TimeServiceClient_GUI_Phase1_Monolithic.cs
	Purpose			Network Time Protocol (NTP) Client
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			February 2015
*/

using System;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;

namespace TimeServiceClient
{
    public partial class TimeServiceClient_GUI_Phase1_Monolithic : Form
    {
        public const Int16 NTP_Port = 123;
        public const Int32 NTP_PACKET_SIZE = 48;// NTP time stamp is in the first 48 bytes of the message
        const Int64 SeventyYears = 2208988800;  // Number of seconds in 70 years. 
                                                // The raw timestamp is the number of seconds Since 1900
                                                // Unix time starts on Jan 1 1970 (70 years = 2208988800 seconds)
	    Socket m_TimeService_Socket;
	    IPEndPoint m_TimeService_IPEndPoint;
        bool m_bNTP_Client_Started;
        Int32 m_iNumRequestsSent;
	    Int32 m_iNumResponsesReceived;

        System.Windows.Forms.Timer m_Timer_SendNTPRequests;
        System.Windows.Forms.Timer m_Timer_WaitForNTPResponses;

        public TimeServiceClient_GUI_Phase1_Monolithic()
        {
            InitializeComponent();
            Populate_NTP_Server_List();
 	        m_bNTP_Client_Started = false;
	        m_iNumRequestsSent = 0;
	        m_iNumResponsesReceived = 0;
            InitialiseControls();
        }

        void InitialiseControls()
        {
            IP_1_textBox.Enabled = false;
            IP_2_textBox.Enabled = false;
            IP_3_textBox.Enabled = false;
            IP_4_textBox.Enabled = false;
            URL_textBox.Enabled = false;
            Port_textBox.Enabled = false;
            ServerURL_listBox.Enabled = true;
            ServerLocation_listBox.Enabled = false;
            UNIX_Time_textBox.Enabled = false;
            UTC_Time_textBox.Enabled = false;
            Num_NTP_Requests_textBox.Enabled = false;
            NUM_NTP_Responses_textBox.Enabled = false;
            Initialise_ServerURL_listBox(); // Selects first item in list boxes, by default
         }

        void Start_button_Click(object sender, EventArgs e)
        {
	        if(false == m_bNTP_Client_Started)
	        {
                Start_button.Text = "STOP NTP requests";
                ServerURL_listBox.Enabled = false;
                m_iNumRequestsSent = 0;
		        m_iNumResponsesReceived = 0;
		        UpdateStatisticsDisplay();
		        StartPeriodic_NTP_requests();
                m_bNTP_Client_Started = true;
            }
	        else
	        {
		        Start_button.Text = "START NTP requests";
                ServerURL_listBox.Enabled = true;
                m_bNTP_Client_Started = false;
		        StopPeriodic_NTP_requests();
	        }
        }

        void StartPeriodic_NTP_requests()
        {
	        // Create Socket and set it NonBlocking
	        try
            {
                m_TimeService_Socket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
	            m_TimeService_Socket.Blocking = false;
            }
            catch
            {
		        MessageBox.Show("Socket create failed","NTP client");
		        CloseSocketAndExit();	
	        }
	
	        Send_TimeService_Request();
            Start_Timer_WaitForNTPResponses();
            Start_Timer_SendNTPRequests();  // Subsequent NTP requests are made every 10 seconds
        }

        void StopPeriodic_NTP_requests()
        {
            StopTimers();
            m_TimeService_Socket.Close();
        }
        
        void Start_Timer_SendNTPRequests()
        {
            m_Timer_SendNTPRequests = new System.Windows.Forms.Timer();
            m_Timer_SendNTPRequests.Interval = 10000; // 10 second period for send NTP request
            m_Timer_SendNTPRequests.Tick += OnTimedEvent_Timer_SendNTPRequests;
            m_Timer_SendNTPRequests.Enabled = true;
        }

        void OnTimedEvent_Timer_SendNTPRequests(Object myObject, EventArgs myEventArgs)
        {
		    Send_TimeService_Request();
        }

        void Start_Timer_WaitForNTPResponses()
        {
            m_Timer_WaitForNTPResponses = new System.Windows.Forms.Timer();
            m_Timer_WaitForNTPResponses.Interval = 200; // 0.2 second period for listening for received NTP responses
            m_Timer_WaitForNTPResponses.Tick += OnTimedEvent_Timer_WaitForNTPResponses;
            m_Timer_WaitForNTPResponses.Enabled = true;
        }

        void OnTimedEvent_Timer_WaitForNTPResponses(Object myObject, EventArgs myEventArgs)
        {
	        DoReceive();
        }

        void DoReceive()
        {
            byte[] bReceivePacketBuffer = new byte [NTP_PACKET_SIZE]; // buffer to hold incoming packets (UTC time value)
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;

            if (null != m_TimeService_Socket && true == m_bNTP_Client_Started)
            {
                iReceiveByteCount = 0;
                try
                {
                    iReceiveByteCount = m_TimeService_Socket.ReceiveFrom(bReceivePacketBuffer, ref From_EndPoint);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {   // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {   // Additional diagnostics can be placed here if necessary
                    }
                    return;
                }
                if (0 < iReceiveByteCount)
                {
 		            // The timestamp starts at byte 40 of the received packet and is four bytes,
                    Int64 l1 = bReceivePacketBuffer[40]; // << 24;
                    Int64 l2 = bReceivePacketBuffer[41]; // << 24;
                    Int64 l3 = bReceivePacketBuffer[42]; // << 24;
                    Int64 l4 = bReceivePacketBuffer[43]; // << 24;
                    Int64 secsSince1900 = (l1 << 24) + (l2 << 16) + (l3 << 8) + l4;  
                    Int64 lUnixTime = secsSince1900 - SeventyYears;	// Subtract seventy years
		            Int32 lHour = (Int32) (lUnixTime  % 86400L) / 3600;
                    Int32 lMinute = (Int32) (lUnixTime % 3600) / 60;
                    Int32 lSecond = (Int32) lUnixTime % 60;

                    UNIX_Time_textBox.Text = lUnixTime.ToString();
                    string sUTC_Time = string.Format("{0:00}:{1:00}:{2:00}", lHour, lMinute, lSecond);
                    UTC_Time_textBox.Text = sUTC_Time;
		            // NTP does not indicate information specifcally related to local time zones such as BST or other daylight savings time offsets
		            m_iNumResponsesReceived++;
		            UpdateStatisticsDisplay();
	            }
            }
        }

        void Send_TimeService_Request()
        {
            byte[] bSendPacketBuffer = new byte [NTP_PACKET_SIZE]; // Zero-out entire 48-byte buffer to hold incoming packets (UTC time value)
            Int32 iSentByteCount;
	        // Initialize values needed to form NTP request
	        bSendPacketBuffer[0] = 0xE3;// 0b11100011;   
								        // LI bits 7,6			= 3 (Clock not synchronised), 
								        // Version bits 5,4,3	= 4 (The current version of NTP)
								        // Mode bits 2,1,0		= 3 (Sent by client)
            try
            {
                iSentByteCount = m_TimeService_Socket.SendTo(bSendPacketBuffer, m_TimeService_IPEndPoint);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   // Additional diagnostics can be placed here if necessary
                }
                return;
            }
	        m_iNumRequestsSent++;
	        UpdateStatisticsDisplay();
        }

        void SetUp_TimeService_AddressStruct(string sURL)
        {
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sURL);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   
                    m_TimeService_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], NTP_Port);
                    
                    // Display the first IPv4 address and the pre-defined port 
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    IP_1_textBox.Text = bAddressBytes[0].ToString();
                    IP_2_textBox.Text = bAddressBytes[1].ToString();
                    IP_3_textBox.Text = bAddressBytes[2].ToString();
                    IP_4_textBox.Text = bAddressBytes[3].ToString();
                    Port_textBox.Text = NTP_Port.ToString();
                    return;
                }
            }
            // No IPv4 address found for the specific URL
            IP_1_textBox.Text = "No";
            IP_2_textBox.Text = "IP";
            IP_3_textBox.Text = "v4";
            IP_4_textBox.Text = "Addr";
        }

        void Populate_NTP_Server_List()
        {
            ServerURL_listBox.Items.Add("time.nist.gov");
            ServerLocation_listBox.Items.Add("NIST round robin load equalisation");
            ServerURL_listBox.Items.Add("time.windows.com");
            ServerLocation_listBox.Items.Add("Windows Time service");
            ServerURL_listBox.Items.Add("nist1-atl.ustiming.org");
	        ServerLocation_listBox.Items.Add("Atlanta, Georgia");
	        ServerURL_listBox.Items.Add("wolfnisttime.com");
	        ServerLocation_listBox.Items.Add("Birmingham, Alabama");
	        ServerURL_listBox.Items.Add("nist1-chi.ustiming.org");
	        ServerLocation_listBox.Items.Add("Chicago, Illinois");
	        ServerURL_listBox.Items.Add("nist1-lnk.binary.net");
	        ServerLocation_listBox.Items.Add("Lincoln, Nebraska");
	        ServerURL_listBox.Items.Add("time-a.timefreq.bldrdoc.gov");
	        ServerLocation_listBox.Items.Add("NIST, Boulder, Colorado");
	        ServerURL_listBox.Items.Add("ntp-nist.ldsbc.edu");
	        ServerLocation_listBox.Items.Add("LDSBC, Salt Lake City, Utah");
	        ServerURL_listBox.Items.Add("nist1-lv.ustiming.org");
	        ServerLocation_listBox.Items.Add("Las Vegas, Nevada");  
	        ServerURL_listBox.Items.Add("nist1-la.ustiming.org");
	        ServerLocation_listBox.Items.Add("Los Angeles, California");
            ServerURL_listBox.Items.Add("nist1-ny.ustiming.org");
            ServerLocation_listBox.Items.Add("New York City, NY");
            ServerURL_listBox.Items.Add("nist1-nj.ustiming.org");
            ServerLocation_listBox.Items.Add("Bridgewater, NJ");
        }

        private void Initialise_ServerURL_listBox()
        {
            ServerURL_listBox.SelectedIndex = 0;
            ServerLocation_listBox.SelectedIndex = 0;
            Get_ServerURL_listBox_Selection();
        }

        private void Get_ServerURL_listBox_Selection()
        {
            string sSelectedURL = ServerURL_listBox.SelectedItem.ToString();
            URL_textBox.Text = sSelectedURL;
            SetUp_TimeService_AddressStruct(sSelectedURL);
        }

        private void ServerURL_listBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            int iSelectionIndex = ServerURL_listBox.SelectedIndex;
            ServerLocation_listBox.SelectedIndex = iSelectionIndex;
            Get_ServerURL_listBox_Selection();
            UNIX_Time_textBox.Text = "";
            UTC_Time_textBox.Text = "";
            m_iNumRequestsSent = 0;
            m_iNumResponsesReceived = 0;
            UpdateStatisticsDisplay();
        }

        void UpdateStatisticsDisplay()
        {
	        Num_NTP_Requests_textBox.Text = m_iNumRequestsSent.ToString();
            NUM_NTP_Responses_textBox.Text = m_iNumResponsesReceived.ToString();
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        void StopTimers()
        {
            if (null != m_Timer_SendNTPRequests)
            {
                m_Timer_SendNTPRequests.Enabled = false;
            }
            if (null != m_Timer_WaitForNTPResponses)
            {
                m_Timer_WaitForNTPResponses.Enabled = false;
            }
        }

        void CloseSocketAndExit()
        {
            StopTimers();
            try
            {
                m_TimeService_Socket.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            System.Environment.Exit(-1);
        }
    }
}
