﻿/*	File			NTP_Client.cs
	Purpose			Network Time Protocol (NTP) Client - Library
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			February 2015
*/

using System;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;

namespace TimeServiceClient_Library
{
    public class NTP_Client
    {
        private const Int16 NTP_Port = 123;
        private const Int32 NTP_PACKET_SIZE = 48;// NTP time stamp is in the first 48 bytes of the message
        private const Int64 SeventyYears = 2208988800;  // Number of seconds in 70 years. 
                                                // The raw timestamp is the number of seconds Since 1900
                                                // Unix time starts on Jan 1 1970 (70 years = 2208988800 seconds)

        public struct NTP_Timestamp_Data
        {		
	        public Int64 lUnixTime;	// Seconds since 1970 (secsSince1900 - seventyYears)
	        public Int32 iHour;
	        public Int32 iMinute;
	        public Int32 iSecond;
	    };

        public enum NTP_Client_ResultCode {NTP_Success = 0, NTP_ServerAddressNotSet = 1, NTP_SendFailed = 2, NTP_ReceiveFailed = 3};

        private bool m_bNTP_Client_Started;
        private bool m_bTimeServiceAddressSet;
	    private Socket m_TimeService_Socket;
	    private IPEndPoint m_TimeService_IPEndPoint;

        public NTP_Client(ref bool bSocketOpenSuccess)
        {	
	        m_bTimeServiceAddressSet = false;
            bSocketOpenSuccess = CreateSocket();
		}

        private bool CreateSocket()
        {
	        // Create Socket and set it NonBlocking
	        try
            {
                m_TimeService_Socket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
	            m_TimeService_Socket.Blocking = false;
            }
            catch
            {
		        return false;	
	        }
	        return true;
        }

        public void SetUp_TimeService_AddressStruct(string sURL, ref string sIP_Address1, ref string sIP_Address2,
                                    ref string sIP_Address3, ref string sIP_Address4, ref string sPort)
        {         
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sURL);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {
                    m_TimeService_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], NTP_Port);
                    m_bTimeServiceAddressSet = true;

                    // Return the first IPv4 address and the pre-defined port 
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    sIP_Address1 = bAddressBytes[0].ToString();
                    sIP_Address2 = bAddressBytes[1].ToString();
                    sIP_Address3 = bAddressBytes[2].ToString();
                    sIP_Address4 = bAddressBytes[3].ToString();
                    sPort = NTP_Port.ToString();
                    return;
                }
            }
            // No IPv4 address found for the specific URL
            sIP_Address1 = "No";
            sIP_Address2 = "IP";
            sIP_Address3 = "v4";
            sIP_Address4 = "Addr";
        }

        public NTP_Client_ResultCode Get_NTP_Timestamp(ref NTP_Timestamp_Data NTP_Timestamp)
        {
	        NTP_Timestamp.iHour = 0;
	        NTP_Timestamp.iMinute = 0;
	        NTP_Timestamp.iSecond = 0;
	        NTP_Timestamp.lUnixTime = 0;

	        if(true == m_bTimeServiceAddressSet)
	        {
		        if(true == Send_TimeService_Request())
		        {   // Send operation succeeded 
			        if(true == Receive(ref NTP_Timestamp))
			        {		
				        return NTP_Client_ResultCode.NTP_Success; // Signal that the NTP_Timestamp has been updated with valid content
			        }
			        return NTP_Client_ResultCode.NTP_ReceiveFailed; // Signal that the receive operation failed (Time server did not reply)
		        }
		        return NTP_Client_ResultCode.NTP_SendFailed; // Signal that the send operation failed (Time server was not contacted)
	        }
            return NTP_Client_ResultCode.NTP_ServerAddressNotSet; // Signal that Time server address has not been set, cannot get NTP timestamp
        }

        private bool Send_TimeService_Request()
        {
            byte[] bSendPacketBuffer = new byte [NTP_PACKET_SIZE]; // Zero-out entire 48-byte buffer to hold incoming packets (UTC time value)
            Int32 iSentByteCount;
	        // Initialize values needed to form NTP request
	        bSendPacketBuffer[0] = 0xE3;// 0b11100011;   
								        // LI bits 7,6			= 3 (Clock not synchronised), 
								        // Version bits 5,4,3	= 4 (The current version of NTP)
								        // Mode bits 2,1,0		= 3 (Sent by client)
            try
            {
                iSentByteCount = m_TimeService_Socket.SendTo(bSendPacketBuffer, m_TimeService_IPEndPoint);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   // Additional diagnostics can be placed here if necessary
                }
                return false;
            }
            return true;
        }

        private bool Receive(ref NTP_Timestamp_Data NTP_Timestamp)
        {
	        System.Threading.Thread.Sleep(500);	// Wait for a short time for the time service response.
	        // In combination with non-blocking receive this prevents the application freezing if the time service does not respond
	        // but waits long enough for the reply RTT so mostly avoids missing an actual reply 
	        // and avoids the need for a timer within the NTP_Client class
	        // Tested with the following values:
	        // 100ms(can work but unreliable) 200ms(generally ok but highly dependent on network RTT) 400ms(generally reliable) 500ms(adds margin of safety)

            byte[] bReceivePacketBuffer = new byte [NTP_PACKET_SIZE]; // buffer to hold incoming packets (UTC time value)
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;

            if (null != m_TimeService_Socket && true == m_bNTP_Client_Started)
            {
                iReceiveByteCount = 0;
                try
                {
                    iReceiveByteCount = m_TimeService_Socket.ReceiveFrom(bReceivePacketBuffer, ref From_EndPoint);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {   // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {   // Additional diagnostics can be placed here if necessary
                    }
                    return false;
                }
                if (0 < iReceiveByteCount)
                {
 		            // The timestamp starts at byte 40 of the received packet and is four bytes,
                    Int64 l1 = bReceivePacketBuffer[40]; // << 24;
                    Int64 l2 = bReceivePacketBuffer[41]; // << 16;
                    Int64 l3 = bReceivePacketBuffer[42]; // << 8;
                    Int64 l4 = bReceivePacketBuffer[43];
                    Int64 secsSince1900 = (l1 << 24) + (l2 << 16) + (l3 << 8) + l4;  
                    NTP_Timestamp.lUnixTime = secsSince1900 - SeventyYears;	// Subtract seventy years
		            NTP_Timestamp.iHour = (Int32) (NTP_Timestamp.lUnixTime  % 86400L) / 3600;
                    NTP_Timestamp.iMinute = (Int32) (NTP_Timestamp.lUnixTime % 3600) / 60;
                    NTP_Timestamp.iSecond = (Int32) NTP_Timestamp.lUnixTime % 60;
	            }
                else
                {
                    return false;
                }
            }
            return true;
        }

        public bool Get_ClientStarted_Flag()
        {
            return m_bNTP_Client_Started;
        }

        public void Set_ClientStarted_Flag(bool bClient_Started)
        {
            m_bNTP_Client_Started = bClient_Started;
        }
        
        public void CloseSocket()
        {
            try
            {
                m_TimeService_Socket.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }
    }
}

